package com.fluentcommerce.common;

import static com.fluentcommerce.jetpack.rubix.utils.rulelogger.aspect.RuleLogAspect.getFlog;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fluentcommerce.jetpack.rubix.utils.rulelogger.annotation.RuleLog;
import com.fluentretail.rubix.rule.meta.RuleInfo;
import com.fluentretail.rubix.v2.context.Context;
import com.fluentretail.rubix.v2.rule.Rule;
import com.fluentretail.se.plugins.util.CommonUtils;
import com.fluentretail.se.plugins.util.Constants;
import com.fluentretail.se.plugins.util.SettingUtils;
import com.fluentretail.se.plugins.util.helpers.LogHelper;
import com.fluentretail.se.plugins.util.helpers.LogHelper.Level;
import com.fluentretail.se.plugins.util.v2.ContextUtils;
import java.util.List;

@RuleLog
public abstract class BaseRule implements Rule {

    private LogHelper logHelper = new LogHelper(Level.INFO, getFlog());

    protected ObjectMapper mapper;

    protected boolean loggingEnabled = true;

    public abstract void execute(ContextWrapper context);

    protected <E extends Exception> void onException(E exception, ContextWrapper context) {
        if (exception instanceof RuntimeException) {
            throw (RuntimeException) exception;
        }
        throw new RuntimeException(exception);
    }

    protected boolean shouldLog() {
        return loggingEnabled;
    }

    protected void persistLogs(ContextWrapper contextWrapper) {
        if (shouldLog()) {
            CommonUtils.newLog(contextWrapper);
        }
    }

    @Override
    public <C extends Context> void run(C context) {
        ContextWrapper contextWrapper = ContextWrapper.from(context);
        logHelper = new LogHelper(
            SettingUtils.getSettingValue(context,
                Constants.LOG_SETTING_PREFIX + getClass().getAnnotation(RuleInfo.class).name(),
                Level.INFO.name()), getFlog());

        try {
            execute(contextWrapper);
        } catch (Exception e) {
            onException(e, contextWrapper);
        } finally {
            persistLogs(contextWrapper);
        }
    }

    protected LogHelper getLog() {
        return logHelper;
    }

    protected ObjectMapper getObjectMapper() {
        if (mapper == null) {
            mapper = new ObjectMapper();
        }
        return mapper;
    }

    protected void debugLog(String message, Object... params) {
        getLog().debug(String.format(message, params));
    }
    protected void infoLog(String message, Object... params) {
        getLog().info(String.format(message, params));
    }

    protected void warnLog(String message, Object... params) {
        getLog().warn(String.format(message, params));
    }

    protected void errorLog(String message, Object... params) {
        getLog().error(String.format(message, params));
    }

    public <T> T getConvertedValue(Context context, String attributeName, TypeReference<T> typeReference) {
        return ContextUtils.getConvertedValue(context, getObjectMapper(), attributeName, typeReference);
    }

    public <T> T getAttribute(Context context, String attributeName, Class clazz) {
        return ContextUtils.getAttribute(context, getObjectMapper(), attributeName, clazz);
    }

    public <T> List<T> getAttributeList(Context context, String attributeName, Class clazz) {
        return ContextUtils.getAttributeList(context, getObjectMapper(), attributeName, clazz);
    }
}
